package cc.smtweb.framework.core.mvc.realm.interceptor;

import cc.smtweb.framework.core.annotation.SwPerm;
import cc.smtweb.framework.core.cache.ISwCache;
import cc.smtweb.framework.core.cache.redis.RedisManager;
import cc.smtweb.framework.core.mvc.controller.IEditor;
import cc.smtweb.framework.core.mvc.realm.exception.ForbiddenException;
import cc.smtweb.framework.core.mvc.realm.exception.UnauthenticatedException;
import cc.smtweb.framework.core.mvc.realm.service.PermCheckItem;
import cc.smtweb.framework.core.mvc.realm.service.PermChecker;
import cc.smtweb.framework.core.session.SessionUtil;
import cc.smtweb.framework.core.session.UserSession;
import org.apache.commons.lang3.StringUtils;

import javax.servlet.http.HttpServletRequest;

public class AbstractPermInterceptor {
    private final RedisManager redisManager;
    private ISwCache<Long, PermChecker> cacheService;

    public AbstractPermInterceptor(RedisManager redisManager) {
        this.redisManager = redisManager;
    }

    public void setCache(ISwCache<Long, PermChecker> cacheService) {
        this.cacheService = cacheService;
    }

    protected boolean handle(HttpServletRequest request, String permissionValue) {
        // 如果注解为null, 说明不需要拦截, 直接放过
        if (SwPerm.NONE.equals(permissionValue)) {
            return true;
        }

        // redis读取session，判断是否登录
        String token = SessionUtil.readToken(request);
        if ("design".equals(token)) return true;
        if (StringUtils.isBlank(token)) {
            throw new UnauthenticatedException("not find Auth-Token in header");
        }

        UserSession us = redisManager.get(token, UserSession.class);
        if (us == null) {
            throw new UnauthenticatedException("not find UserSession by token: " + token);
        }

        request.setAttribute(IEditor.USER_TOKEN, token);
        request.setAttribute(IEditor.USER_SESSION, us);

        // 如果标记了权限注解，则判断权限
        if (checkPermission(permissionValue, us)) {
            // 更新Token redis TTL
            redisManager.expire(token, RedisManager.SESSION_EXPIRE_SEC);
            return true;
        } else {
            throw new ForbiddenException("user not permission: " + permissionValue);
        }
    }

    /**
     * 权限检查
     */
    private boolean checkPermission(String permissionValue, UserSession us) {
        if (StringUtils.isBlank(permissionValue)) {
            return true;
        }

        // 从本地缓存或数据库中获取该用户的权限信息
        PermChecker permissionSet = cacheService.get(us.getUserId());

//    if (MapUtils.isEmpty(permissionSet)) {
//      throw new ForbiddenException("empty permission");
//    }

        PermCheckItem permChecker = permissionSet.get(permissionValue);

        if (permChecker != null) {
            return true;
        }

        while (true) {
            permissionValue = getParentPermValue(permissionValue);

            if (permissionValue != null) {
                permChecker = permissionSet.get(permissionValue);
                if (permChecker != null && permChecker.isPerfixMath()) {
                    return true;
                }
            } else {
                break;
            }
        }

        return false;
    }

    private static String getParentPermValue(String permissionValue) {
        if (permissionValue.length() > 0) {

            int pos = permissionValue.lastIndexOf(':');
            if (pos > 0) {
                return permissionValue.substring(0, pos);
            }

            return "";
        }

        return null;
    }
}
